import re
import numpy as np

from .periodic_table import *
from ..units import Quantity, Unit
from .. import RowCollector, ParameterTable

PERIODIC_TABLE = ParameterTable(PT_HEADER, PT_DATA, keys=True)

class Element:
    expression: str
    count: int
    natural: bool
    element: str
    isotope: int
    ionisation: int
    A: Quantity = None
    Z: float = None
    N: float = None
    e: float = None
    n: Quantity = None
    rho: Quantity = None

    # get a specific isotope
    def get_isotope(self, element:str, isotope: int, ionisation: int):
        isotopes = PERIODIC_TABLE[element]
        Z = isotopes.Z
        iso = isotope if isotope else Z*2
        ion = ionisation if ionisation else 0
        if str(iso) not in isotopes.A:
            raise Exception("Required isotope could not be found:", element, iso)
        M, NA = isotopes.A[str(iso)]
        A = Quantity(M, 'Da') + Quantity(ion, '[m_e]')
        N = int(iso-isotopes.Z)
        e = isotopes.Z+ion
        return NA, A, Z, N, e, iso, ion
        
    # get the most abundant isotope
    def get_abundant(self, element:str, ionisation:int):
        isotopes = PERIODIC_TABLE[element]
        idmax = np.argmax(iso[1] for iso in isotopes.A.values())
        isotope = int(list(A for A in isotopes.A.keys())[idmax])
        return self.get_isotope(element, isotope, ionisation)        

    # get a natural isotopic average
    def get_natural(self, element:str, ionisation:int):
        isotopes = PERIODIC_TABLE[element]
        with RowCollector(['NA', 'A', 'Z', 'N', 'e', 'iso', 'ion']) as rc:
            for iso in isotopes.A.keys():
                rc.append(self.get_isotope(element, int(iso), ionisation))
            return (
                np.sum(rc.NA), 
                np.average(rc.A, weights=rc.NA),
                np.average(rc.Z, weights=rc.NA),
                np.average(rc.N, weights=rc.NA),
                np.average(rc.e, weights=rc.NA),
                np.average(rc.iso, weights=rc.NA),
                int(rc.ion[0]),
            )

    def __init__(self, expression:str, count:int=1, natural:bool=True):
        self.expression = expression
        self.count = count
        self.natural = natural
        # parse the expression
        if m := re.match("(\[(p|n|e)\])", expression):
            nucleon = m.group(2)
            self.element = expression
            NA, iso, ion = 100.0, None, None
            A = Unit(f"[m_{nucleon}]").to('Da')
            nucleons = {'p':(1,0,0),'n':(0,1,0),'e':(0,0,1)}
            Z, N, e = nucleons[nucleon]
            self.isotope = 0
            self.ionisation = 0
        elif m := re.match("([a-zA-Z]{1,2})(\{([0-9]+)([+-]{1}[0-9]*)\}|\{([0-9]+)\}|\{([+-]{1}[0-9]*)\}|)", expression):
            # Extract information about isos
            element, variant, iso1, ion1, iso2, ion3 = m.groups()
            if element=='D':
                element, variant, iso1, ion1, iso2, ion3 = 'H', True, 2, ion1, 2, None
            elif element=='T':
                element, variant, iso1, ion1, iso2, ion3 = 'H', True, 3, ion1, 3, None
            self.element = element
            if iso1 and ion1:
                if ion1=="-": ion1="-1"
                elif ion1=="+": ion1="1"
                self.isotope, self.ionisation = int(iso1), int(ion1)
            elif iso2:
                self.isotope, self.ionisation = int(iso2), 0
            elif ion3:
                if ion3=="-": ion3="-1"
                elif ion3=="+": ion3="1"
                self.isotope, self.ionisation = None, int(ion3)
            else:
                self.isotope, self.ionisation = None, 0
            # set element values
            if self.isotope:
                NA, A, Z, N, e, iso, ion = self.get_isotope(self.element, self.isotope, self.ionisation)
            elif self.natural:
                NA, A, Z, N, e, self.isotope, self.ionisation = self.get_natural(self.element, self.ionisation)
            else:
                NA, A, Z, N, e, self.isotope, self.ionisation = self.get_abundant(self.element, self.ionisation)
        else:
            raise Exception('Unrecognized expression', expression)
        self.A = self.count*A
        self.Z = self.count*Z
        self.N = self.count*N
        self.e = self.count*e
        
    def __mul__(self, other:float):
        return Element(self.expression, self.count*other, natural=self.natural)
            
    def __add__(self, other:'Element'):
        if self.expression!=other.expression:
            raise Exception("Only same elements can be added up:", self.expression, other.expression)
        return Element(self.expression, self.count+other.count, natural=self.natural)
        
    def __str__(self):
        if self.count>1:
            return f"Element({self.expression}{self.count} Z={self.Z} N={self.N:.3f} e={self.e} A={self.A.value('Da'):.3f})"
        else:
            return f"Element({self.expression} Z={self.Z} N={self.N:.3f} e={self.e} A={self.A.value('Da'):.3f})"
        
    def set_density(self, n:Quantity):
        self.n = self.count * n
        self.rho = self.A * n
            


import numpy as np
import re
from math import isclose
import copy

from .element import Element
from .material_solver import MaterialSolver
from .. import ParameterTable, RowCollector
from ..units import Quantity, Unit

class Molecule:
    natural: bool
    elements: dict
    M: Quantity
    rho: Quantity = None
    V: Quantity = None
    n: Quantity = None
    
    @staticmethod
    def from_elements(elements:list, natural:bool=True):
        molecule = Molecule(natural=natural)
        for el in elements:
            molecule.add_element(el)
        return molecule
        
    def atom(self, expression:str):
        if m:=re.match("[0-9]+(\.[0-9]+|)([eE]{1}[+-]?[0-9]{0,3}|)",expression):
            return float(expression)
        else:
            return Molecule.from_elements([
                Element(expression, natural=self.natural),
            ], natural=self.natural)
            
    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass
            
    def __init__(self, expression:str=None, natural:bool=True):
        self.natural = natural
        self.elements = {}
        if expression and expression!='':
            with MaterialSolver(self.atom) as ms:
                molecule = ms.solve(expression)
            for expr, el in molecule.elements.items():
                self.elements[expr] = el
        self.M = np.sum([e.A for e in self.elements.values()])

    def __str__(self):
        elements = []
        for expr, el in self.elements.items():
            element = f"{expr}" 
            if el.count>1:
                element += f"{int(el.count):d}"
            elements.append(element)
        elements = " ".join(elements)
        data = self.data_molecule(quantity=False)
        p = int(round(data['sum']['Z']))
        n = data['sum']['N']
        e = int(round(data['sum']['e']))
        A = data['sum']['A']
        return f"Molecule(p={p:d} n={n:.03f} e={e:d} A={A:.03f})"
        
    def __mul__(self, other:float):
        molecule = copy.deepcopy(self)
        molecule.elements = {}
        for expr in self.elements.keys():
            el = self.elements[expr] * other
            molecule.add_element(el)
        return molecule
    
    def __add__(self, other):
        molecule = copy.deepcopy(self)
        molecule.elements = {}
        for expr,el in self.elements.items():
            molecule.add_element(el)
        if isinstance(other, Molecule):
            for expr,el in other.elements.items():
                molecule.add_element(el)
        elif isinstance(other, Element):
            molecule.add_element(other)
        return molecule
    
    def add_element(self, element:Element):
        expr = element.expression
        if self.rho:
            element.set_density(self.n)
        if expr in self.elements:
            self.elements[expr] += element
        else:
            self.elements[expr] = element
        self.M = np.sum([e.A for e in self.elements.values()])
    
    def set_amount(self, rho:Quantity, V:Quantity=None):
        self.rho = rho
        self.n = (rho/self.M).rebase()
        for s in self.elements.keys():
            self.elements[s].set_density(self.n)
        self.V = V
            
    def data_elements(self, quantity=True):
        with ParameterTable(['element','isotope','ionisation','A','Z','N','e'], #,'A_nuc','E_bin'], 
                keys=True, keyname='expression') as pt:
            for s,e in self.elements.items():
                el = Element(s, natural=self.natural)
                #A_nuc = Quantity(el.Z, '[m_p]') + Quantity(el.N, '[m_n]')
                #E_bin = ((A_nuc-el.A)*Unit('[c]')**2)/(el.Z+el.N)
                pt[s] = [
                    el.element, 
                    el.isotope, 
                    el.ionisation, 
                    el.A.to('Da') if quantity else el.A.value('Da'), 
                    el.Z, 
                    el.N, 
                    el.e, 
                    #A_nuc.value('Da'), 
                    #E_bin.value('MeV'),
                ]
            return pt
            
    def data_molecule(self, part:list=None, quantity=True):
        if not self.elements:
            return None
        columns = ['A','Z','N','e']
        if self.rho:
            columns += ['n','rho','X']
        if self.V:
            columns += ['n_V','M_V']
        with ParameterTable(['count']+columns, keys=True, keyname='expression') as pt:
            rc = RowCollector(['count']+columns)
            for s,e in self.elements.items():
                if part and s not in part:
                    continue
                row = [
                    e.count, 
                    e.A.to('Da') if quantity else e.A.value('Da'), 
                    e.Z, 
                    e.N, 
                    e.e, 
                ]
                if self.rho:
                    row += [
                        e.n.to('cm-3') if quantity else e.n.value('cm-3'), 
                        e.rho.to('g/cm3') if quantity else e.rho.value('g/cm3'),
                        (e.rho/self.rho).to('%') if quantity else (e.rho/self.rho).value('%'),
                    ]
                if self.V:
                    row += [
                        (e.n*self.V).value(),
                        (e.rho*self.V).to('g') if quantity else (e.rho*self.V).value('g'),
                    ]
                pt[s] = row
                rc.append(row)
            avg = [np.average(rc['count'])]
            sum = [np.sum(rc['count'])]
            for p in columns:
                sum.append(np.sum(rc[p]))
                avg.append(np.average(np.divide(rc[p],rc['count']), weights=rc['count']))
            pt['avg'] = avg
            pt['sum'] = sum
            return pt
    
    def print_elements(self):
        df = self.data_elements(quantity=False).to_dataframe()
        df = df.rename(columns={"A": "A[Da]", "A_nuc": "A_nuc[Da]", "E_bin": "E_bin[MeV]"})
        print( df.to_string(index=False) )

    def print_molecule(self, part:list=None):
        df = self.data_molecule(part=part, quantity=False).to_dataframe()
        columns = {"A": "A[Da]"}
        if self.rho:
            columns.update({"n": "n[cm-3]", "rho": "rho[g/cm3]", "X": "X[%]"})
        if self.V:
            columns.update({"M_V": "M_V[g]"})
        df = df.rename(columns=columns)
        print( df.to_string(index=False) )
        
    def print(self):
        print("Properties:")
        print()
        print(f"Molecular mass: {self.M}")
        if self.rho:
            print(f"Mass density: {self.rho}")
            print(f"Molecular density: {self.n}")
        if self.V:
            print(f"Volume: {self.V}")
        print("")
        print("Elements:")
        print("")
        self.print_elements()
        print("")
        print("Molecule:")
        print("")
        self.print_molecule()


import numpy as np
import pytest
from math import isclose
import os
import sys
sys.path.insert(0, 'src')

from scinumtools.units import Quantity, Unit 
from scinumtools.materials import *

def test_expression():
    
    # Natural abundance
    e = Element('B')
    assert str([e.A,e.Z,e.N,e.e]) == "[Quantity(1.081e+01 Da), 5.0, 5.801, 5.0]"
    assert isclose(e.A.value(), 10.811028046410001, rel_tol=MAGNITUDE_PRECISION)
    
    # Isotope
    e = Element('B{11}')
    assert str([e.A,e.Z,e.N,e.e]) == "[Quantity(1.101e+01 Da), 5, 6, 5]"
    assert isclose(e.A.value(), 11.009306, rel_tol=MAGNITUDE_PRECISION)

    # Isotope and ionisation
    e = Element('B{11-1}')
    assert str([e.A,e.Z,e.N,e.e]) == "[Quantity(1.101e+01 Da), 5, 6, 4]"
    assert isclose(e.A.value(), 11.008757420144217, rel_tol=MAGNITUDE_PRECISION)

def test_print():
    
    e = Element('O')
    assert str(e) == "Element(O Z=8.0 N=8.004 e=8.0 A=15.999)"
    assert str((e.Z, e.N, e.e, e.A)) == "(8.0, 8.00448, 8.0, Quantity(1.600e+01 Da))"

    e = Element('O', natural=False)
    assert str(e) == "Element(O Z=8 N=8.000 e=8 A=15.995)"
    
    e = Element('O', count=2, natural=False)
    assert str(e) == "Element(O2 Z=16 N=16.000 e=16 A=31.990)"
    assert e.count == 2
    assert e.element == 'O'
    assert e.isotope == 16
    assert e.ionisation == 0
    
    e = Element('O{17-2}')
    assert str(e) == "Element(O{17-2} Z=8 N=9.000 e=6 A=16.998)"

def test_arithmetic():
    
    e1 = Element('B')
    e2 = Element('B',2)

    # addition
    e12 = e1 + e2
    assert e12.count == e1.count + e2.count
    assert e12.A == e1.A + e2.A
    assert e12.Z == e1.Z + e2.Z
    assert e12.N == e1.N + e2.N
    assert e12.e == e1.e + e2.e

    # multiplication
    e13 = e1 * 3
    assert e13.count == e1.count * 3
    assert e13.A == e1.A * 3
    assert e13.Z == e1.Z * 3
    assert e13.N == e1.N * 3
    assert e13.e == e1.e * 3
